from neo4j import GraphDatabase
from .graph import Graph
from .algorithms import (
    AlgoWrite,
    AlgoStream,
    AlgoStats,
    Function
)


class ItemGetter:

    def __init__(self, driver, namespace=None):
        self.driver = driver
        self.namespace = namespace

    def __call__(self, *args, **kwargs):
        return Function(self.driver, self.namespace)(*args, **kwargs)

    @property
    def graph(self, *args, **kwargs):
        return Graph(self.driver, self.namespace)

    @property
    def write(self, *args, **kwargs):
        return AlgoWrite(self.driver, self.namespace)

    @property
    def stream(self, *args, **kwargs):
        return AlgoStream(self.driver, self.namespace)

    @property
    def stats(self, *args, **kwargs):
        return AlgoStats(self.driver, self.namespace)

    def __getattr__(self, attr):
        namespace = ".".join([self.namespace, attr])
        return ItemGetter(self.driver, namespace)


class GDS:

    namespace = "gds"

    def __init__(self, uri, auth):
        self.uri = uri
        self.auth = auth
        self.driver = GraphDatabase.driver(self.uri, auth=self.auth)

    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        self.driver.close()

    def __del__(self):
        try:
            self.driver.close()
        except Exception:
            pass

    def __getattr__(self, attr):
        return getattr(ItemGetter(self.driver, self.namespace), attr)




class Neo4jRunner:
    def run_cypher(self, cypher, params, get_data=True):
        with self.driver.session() as s:
            r = s.run(cypher, params)
        if get_data:
            data = r.data()
            return data
        return r



from .mixins import Neo4jRunner


class GraphCreate(Neo4jRunner):

    def __init__(self, driver, namespace):
        super().__init__()
        self.namespace = namespace
        self.driver = driver

    def __call__(self, graph_name, node_projection="*", relationship_projection="*"):
        """Create a projected graph from native projection
        """
        cypher = "CALL gds.graph.create($graph_name, $node_projection, $relationship_projection)"
        return self.run_cypher(cypher, {
            "graph_name": graph_name,
            "node_projection": node_projection,
            "relationship_projection": relationship_projection
        })

    def cypher(self, graph_name, node_projection, relationship_projection):
        """Create a projected graph from Cypher projection
        """
        cypher = "CALL gds.graph.create.cypher($graph_name, $node_projection, $relationship_projection)"
        return self.run_cypher(cypher, {
            "graph_name": graph_name,
            "node_projection": node_projection,
            "relationship_projection": relationship_projection
        })


class Graph(Neo4jRunner):

    def __init__(self, driver, namespace):
        super().__init__()
        self.namespace = namespace
        self.driver = driver

    @property
    def create(self, *args, **kwargs):
        return GraphCreate(self.driver, self.namespace, *args, **kwargs)

    def drop(self, graph_name):
        """Drop an existing projected graph
        """
        cypher = "CALL gds.graph.drop($graph_name)"
        return self.run_cypher(cypher, {
            "graph_name": graph_name,
        })



from .mixins import Neo4jRunner


class Function(Neo4jRunner):

    def __init__(self, driver, namespace):
        super().__init__()
        self.namespace = namespace
        self.driver = driver

    def __call__(self, *args):
        """Run algorithm and write results back into Neo4j
        Returns some stats about algorithm execution
        """
        func_name = self.namespace
        params = {str(k): a for k, a in enumerate(args)}
        params_names = params.keys()
        cypher = f"RETURN {func_name}($" + ", $".join(params_names) + ") as result"
        return self.run_cypher(cypher, params, get_data=False).single().get("result")


class EstimateMixin:

    def estimate(self, graph_name, algo_config=None):
        """Run algorithm and write results back into Neo4j
        Returns some stats about algorithm execution
        """
        proc_name = self.procedure_name + ".estimate"
        cypher = f"CALL {proc_name}($graph_name, $algo_config)"
        return self.run_cypher(cypher, {
            "graph_name": graph_name,
            "algo_config": algo_config,
        })


class AlgoMixin(EstimateMixin, Neo4jRunner):

    return_data = True

    def __init__(self, driver, namespace):
        super().__init__()
        self.namespace = namespace
        self.driver = driver

    def _make_params(self, graph_name, algo_config=None):
        return {
            "graph_name": graph_name,
            "algo_config": algo_config,
        }

    def __call__(self, graph_name, algo_config=None):
        cypher = f"CALL {self.procedure_name}($graph_name, $algo_config)"
        return self.run_cypher(
            cypher,
            self._make_params(graph_name, algo_config), self.return_data
        )

    @property
    def procedure_name(self):
        return self.namespace + "." + self.suffix


class AlgoWrite(AlgoMixin):
    suffix = "write"


class AlgoStream(AlgoMixin):
    suffix = "stream"
    return_data = False


class AlgoStats(AlgoMixin):
    suffix = "stats"



import setuptools


setuptools.setup(
    name="pygds",
    version="0.1.0",
    author="Estelle Scifo",
    description="A python wrapper around the Neo4j Graph Data Science Library",
    url="https://github.com/stellasia/pygds",
    packages=["pygds"],
    python_requires='>=3.6',
    install_requires=["neo4j"]
)
